(************************************************)
(*   Turbo Vision Library                       *)
(*   DataBase File Viewer & Editor              *)
(*   ࠡ稪 - 㤪  ਣ쥢    *)
(*   Nick Doodka, addr - 2:4641/58.18@FidoNet   *)
(************************************************)

unit DbEdit;

interface

uses Dialogs, Drivers, Objects, Views, Fox;

type
  PCharCollection = ^TCharCollection;
  TCharCollection = object(TCollection)
    Overflow: Boolean;
    procedure Error(Code, Info: Integer); virtual;
    procedure FreeItem(Item: Pointer); virtual;
  end;

  PHeaderView = ^THeaderView;                   {   }
  THeaderView = object(TView)
    Light: Boolean;
    constructor Init(var Bounds: TRect);
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure Draw; virtual;
  end;

  PFooterView = ^TFooterView;                   { 稪 ᥩ }
  TFooterView = object(TView)
    constructor Init(var Bounds: TRect);
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure Draw; virtual;
  end;

  PEditLine = ^TEditLine;                       { ப   }
  TEditLine = object(TInputLine)
    Original: PString;
    constructor Init;
    destructor Done; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure SetData(var Rec); virtual;
  end;

  PBaseView = ^TBaseView;
  TBaseView = object(TListBox)
    Base: PBase;                                {   }
    Fields: PStringCollection;                  {   ᬮ }
    Margin: Integer;                            {  ࠭ }
    Column: Integer;                            { ⥪  }
    destructor Done; virtual;
    procedure FocusItem(Item: Integer); virtual;
    function GetText(Item, MaxLen: Integer): string; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    function GetPalette: PPalette; virtual;
    procedure SetState(AState: Word; Enable: Boolean); virtual;
    procedure MakeList(ABase: PBase; AFields: PSItem);
    function NewLine: PChar;
  private
    Offset: Word;                               { ᬥ饭  砫  }
    function Width(Name: string): Integer;      { ਭ   ⮬ }
  end;                                          { ஡ ᫥ / ࠢ }

  PDataViewer = ^TDataViewer;          { ꥪ - ᬮ   }
  TDataViewer = object(TGroup)
    Listing: PBaseView;
    HScroll: PScrollBar;
    EdLine: PEditLine;
    constructor Init(var R: TRect; PDbf: PBase; PFields: PSItem);
    procedure Refresh;                          {  ⥪  }
    procedure Reload;                           { १㧨 ᯨ᮪ }
  end;

  PDataEditor = ^TDataEditor;          { ꥪ - ।   }
  TDataEditor = object(TDataViewer)
    JumpVert: Boolean; {⨪쭮 / ਧ⠫쭮 ६饭}
    constructor Init(var R: TRect; PDbf: PBase; PFields: PSItem);
    procedure HandleEvent(var Event: TEvent); virtual;
  private
    procedure AutoSave;
  end;

  PCellEditor = ^TCellEditor;          { ꥪ - ।   }
  TCellEditor = object(TDataEditor)
    constructor Init(var R: TRect; PDbf: PBase; PFields: PSItem);
    procedure HandleEvent(var Event: TEvent); virtual;
  end;

implementation {============================================================}

{          Tab=RecNo        Data from *.dbf           }
{ PChar ->  }

uses Strings, Memory, MsgBox, Lib;

const
  Tab       =  10;   {      }
  Delimiter = '';   { ᨬ-ࠧ⥫ ᥩ }

  cmFocused = 1900;  { ࠭         }
  cmColumn  = 1901;  { ࠭        }
  cmMargin  = 1902;  {  ࠭    }
  cmResetLn = 1903;  { .⠭ ப  }

procedure TCharCollection.Error(Code, Info: Integer);
begin
  if Code = coOverflow then
    Overflow := True
  else
    inherited Error(Code, Info);
end;

procedure TCharCollection.FreeItem(Item: Pointer);
begin
  StrDispose(Item);
end;

{ *** THeaderView object *** }

constructor THeaderView.Init(var Bounds: TRect);
begin
  inherited Init(Bounds);
  EventMask := EventMask or evBroadcast;
end;

procedure THeaderView.HandleEvent(var Event: TEvent);
begin
  inherited HandleEvent(Event);
  if Event.What = evBroadcast then
  case Event.Command of
    cmMargin: DrawView;
    cmReceivedFocus:
      if Event.InfoPtr = Owner then
      begin
        Light := True;
        DrawView;
        ClearEvent(Event);
      end;
    cmReleasedFocus:
      if Event.InfoPtr = Owner then
      begin
        Light := False;
        DrawView;
        ClearEvent(Event);
      end;
  end; {case}
end;

procedure THeaderView.Draw;
var
  S: string[MaxViewWidth];
  Name: string[10];
  Buf: TDrawBuffer;
  Color: Byte;
  I: Integer;
begin
  S := '';
  with PDataViewer(Owner)^.Listing^ do
  for I := Margin to Fields^.Count - 1 do
  begin
    Name := string(Fields^.At(I)^);
    Color := Width(Name);
    if Color > Self.Size.X then Color := Self.Size.X;
    S := S + PadCh(Name, '', Color) + '';
    if Length(S) > Self.Size.X then Break;
  end;
  S := Copy(S, 1, Length(S) - 1) + '';
  if Light then Color := GetColor(8) else Color := GetColor(7);
  MoveChar(Buf, ' ', Color, Size.X);
  MoveStr(Buf[1], S, Color);
  WriteLine(0, 0, Size.X, 1, Buf);
end;

{ *** TFooterView object *** }

constructor TFooterView.Init(var Bounds: TRect);
begin
  inherited Init(Bounds);
  EventMask := EventMask or evBroadcast;
end;

procedure TFooterView.HandleEvent(var Event: TEvent);
begin
  inherited HandleEvent(Event);
  if (Event.What = evBroadcast) and
    (Event.Command = cmFocused) then DrawView;
end;

procedure TFooterView.Draw;
var
  Params: array[0..2] of Longint;
  Display: string[20];
  InfStr: string[3];
  B: TDrawBuffer;
  Color: Byte;
begin
  InfStr := '';
  with PDataViewer(Owner)^.Listing^.Base^ do
  begin
    Params[0] := RecNo;
    Params[1] := LastRec;
    if Deleted then InfStr := 'DEL';
  end;
  if Params[0] > Params[1] then InfStr := 'NEW';
  Params[2] := LongInt(@InfStr);
  FormatStr(Display, ' %d/%d %s', Params);
  Color := GetColor(8);
  MoveChar(B, ' ', Color, Size.X);
  MoveStr(B, Display, Color);
  WriteLine(0, 0, Size.X, 1, B);
end;

{ *** TEditLine object *** }

constructor TEditLine.Init;
var
  R: TRect;
begin
  R.Assign(0, 0, 0, 0);
  inherited Init(R, 255);
  EventMask := EventMask or evBroadcast;
end;

destructor TEditLine.Done;
begin
  DisposeStr(Original);
  inherited Done;
end;

procedure TEditLine.SetData(var Rec);
begin
  inherited SetData(Rec);
  DisposeStr(Original);
  Original := NewStr(Data^);
end;

procedure TEditLine.HandleEvent(var Event: TEvent);

  procedure LocateLine(Link: PBaseView; var Moved: Boolean);
  var
    R: TRect;
    S: string;
    I, Len: Integer;
    Value: Integer;
  begin
    with Link^ do
    begin
      GetBounds(R);
      R.A.X := R.A.X + 1;
      R.A.Y := R.A.Y + Focused - TopItem;
      Len := 0;  Value := Margin;
      if Column > Margin then
      begin
        for i := Value to Column do
          Inc(Len, Width(PString(Fields^.At(I))^) + 1);
        while (Value < Column) and (Len > Size.X) do
        begin
          Dec(Len, Width(PString(Fields^.At(Value))^) + 1);
          Inc(Value);
        end;
      end else
        Value := Column;
      Moved := Value <> Margin;
      Margin := Value;
      for i := Value to Column - 1 do
        Inc(R.A.X, Width(PString(Fields^.At(I))^) + 1);
      S := PString(Fields^.At(Column))^;
      R.B.X := R.A.X + Width(S);
      if R.B.X > Size.X then R.B.X := Size.X;
      R.B.Y := R.A.Y + 1;
      S := Base^.Get(S);
    end;
    SetData(S);
    Locate(R);
  end;

var
  MarginChanged: Boolean;
  P: PBaseView;

begin  { of proc TEditLine.HandleEvent }
  P := PDataViewer(Owner)^.Listing;
  if (Event.What = evKeyDown) and (Event.CharCode in [#32..#255])
  and (Length(Data^) >= P^.Base^.FSize(PString(P^.Fields^.At(P^.Column))^))
  and ((Original = nil) or (Original <> nil) and (Original^ <> Data^)) then
  begin
    Beep(1000, 100);
    ClearEvent(Event);
  end;

  inherited HandleEvent(Event);
  if (Event.What = evBroadcast) and ((Event.Command = cmColumn) or
    (Event.Command = cmFocused) or (Event.Command = cmResetLn)) then
  begin
    LocateLine(Event.InfoPtr, MarginChanged);
    if MarginChanged then Message(Owner, evBroadcast, cmMargin, @Self);
  end;
end;

{ *** TBaseView object *** }

destructor TBaseView.Done;
begin
  if Assigned(Base) and Base^.Modified then Base^.Flush;
  MakeList(nil, nil);
  inherited Done;
end;

procedure TBaseView.MakeList(ABase: PBase; AFields: PSItem);
var
  Q: PSItem;
  I: Integer;
  Temp: PCharCollection;
begin
  if List <> nil then Dispose(List, Done);  List := nil;
  if Fields <> nil then Dispose(Fields, Done);  Fields := nil;
  Base := ABase;  if Base = nil then Exit;
  Margin := 0;  Offset := Tab;
  Fields := New(PStringCollection, Init(Base^.FCount, 5));
  Temp := New(PCharCollection, Init(Base^.LastRec + 1, 50));
  if AFields = nil then  { 㧪     }
    for I := Base^.FCount downto 1 do
      Fields^.AtInsert(0, NewStr(Base^.Field(I)))
  else
    while AFields <> nil do
    begin
      Q := AFields;
      if Base^.AtF(Q^.Value^) <> nil then
        Fields^.AtInsert(Fields^.Count, NewStr(Upper(Q^.Value^)));
      AFields := AFields^.Next;
      DisposeStr(Q^.Value);
      Dispose(Q);
    end;
  Base^.LinkOff;  { ⪫ ࠡ ⭮襭 }
  Base^.GoTop;
  while not Base^.Eof do
  begin
    Temp^.Insert(NewLine);
    if LowMemory or Temp^.Overflow then
    begin
      Beep(600, 150);
      MessageBox(^C'筮 ⨢   㧪  '+
        'ᥩ   %S', @Base^.Dbf, mfWarning + mfOKButton);
      Break;
    end; {if}
    Base^.Skip(1);
  end; {while}
  Base^.LinkOn;  {  ࠡ ⭮襭 }
  Temp^.Insert(nil);
  NewList(Temp);
end;

function TBaseView.NewLine: PChar;
var
  P: PChar;
  BufLen: Word;
  S: array[0..$FF] of Char;

  procedure IncBufLen(Q: PString); far;
  begin
    Inc(BufLen, Width(Q^) + Length(Delimiter));
  end;

  procedure StrConcat(Q: PString); far;
  begin
    P := StrCat(StrCat(P, StrPCopy(S,
      RPad(' ' + Base^.Get(Q^), Width(Q^)))), Delimiter);
  end;

begin { func.NewLine }
  BufLen := Tab + 1;
  Fields^.ForEach(@IncBufLen);
  P := MAlloc(BufLen);
  Str(Base^.RecNo:Tab, S);
  StrCat(P, S);
  Fields^.ForEach(@StrConcat);
  NewLine := P;
end;

procedure TBaseView.FocusItem(Item: Integer);
var
  P: PChar;
  Code: Integer;
  Current: Longint;
  B: array[0..Tab] of Char;
begin
  inherited FocusItem(Item);
  P := List^.At(Item);
  if P = nil then Current := 0 else
    Val(Copy(StrPas(StrLCopy(B, P, Tab)), 1, Tab), Current, Code);
  Base^.GoRec(Current);
  Message(Owner, evBroadcast, cmFocused, @Self);
end;

function TBaseView.GetText(Item, MaxLen: Integer): string;
var
  P: PChar;
  B: array[0..$FF] of Char;
begin
  P := List^.At(Item);
  if P = nil then GetText := '' else
    GetText := StrPas(StrLCopy(B, P + Offset, MaxLen));
end;

procedure TBaseView.HandleEvent(var Event: TEvent);

  procedure CalcOffset;
  var
    I: Integer;
    Name: string[10];
  begin
    Offset := Tab;
    for I := 0 to Margin - 1 do
    begin
      Name := string(Fields^.At(I)^);
      Inc(Offset, Width(Name) + Length(Delimiter));
    end;
  end;

begin
  inherited HandleEvent(Event);
  if Event.What = evBroadcast then
  case Event.Command of
    cmScrollBarChanged:
      if Event.InfoPtr = PDataViewer(Owner)^.HScroll then
      begin
        Column := PScrollBar(Event.InfoPtr)^.Value;
        Message(Owner, evBroadcast, cmColumn, @Self);
        ClearEvent(Event);
      end;
    cmMargin: {   ࠭ }
      begin
        CalcOffset;
        DrawView;
        ClearEvent(Event);
      end;
  end;
end;

procedure TBaseView.SetState(AState: Word; Enable: Boolean);
begin
  inherited SetState(AState, Enable);
  if (AState = sfActive) and Enable then
    Message(Owner, evBroadcast, cmResetLn, @Self);
end;

function TBaseView.Width(Name: string): Integer;
var
  W: Integer;
begin
  W := Base^.FSize(Name) + 2;
  if Length(Name) > W then W := Length(Name);
  Width := W;
end;

function TBaseView.GetPalette: PPalette;
const
  P: string[Length(CListViewer)] = CListViewer;
begin
  P[3] := #26;
  P[4] := #26;
  GetPalette := @P;
end;

{ *** TDataViewer object *** }

constructor TDataViewer.Init(var R: TRect; PDbf: PBase; PFields: PSItem);
var
  View: PScrollBar;
  Step: Integer;
begin
  inherited Init(R);
  Options := Options or ofFramed;

{Vertical ScrollBar}
  R.Assign(Size.X-1, 1, Size.X, Size.Y-1);
  View := New(PScrollBar, Init(R));
  Insert(View);

{Listing of Records}
  R.Assign(0, 1, Size.X-1, Size.Y-1);
  Listing := New(PBaseView, Init(R, 1, View));
  Listing^.MakeList(PDbf, PFields);
  Insert(Listing);

{Header}
  R.Assign(0, 0, Size.X, 1);
  Insert(New(PHeaderView, Init(R)));

{Footer}
  R.Assign(0, Size.Y - 1, Size.X, Size.Y);
  Insert(New(PFooterView, Init(R)));

{Horizontal ScrollBar}
  R.Assign(17, Size.Y-1, Size.X-1, Size.Y);
  HScroll := New(PScrollBar, Init(R));
  with HScroll^ do
  begin
    Options := Options or ofPostProcess;
    Step := Listing^.Fields^.Count - 1;
    SetParams(0, 0, Step, Step, 1);
  end;
  Insert(HScroll);

{EditLine}
  EdLine := New(PEditLine, Init);
  with EdLine^ do Options := Options and not ofSelectable;
  Insert(EdLine);

  SelectNext(False);
end;

procedure TDataViewer.Refresh;
var
  C: PCollection;
  I: Integer;
begin
  Lock;
  with Listing^ do
  begin
    I := Focused;
    List^.AtPut(Focused, NewLine);
    if Focused = List^.Count - 1 then
    begin
      List^.Insert(nil);
      C := List;
      List := nil;
      NewList(C);
    end;
    FocusItem(I);
  end;
  Unlock;
end;

procedure TDataViewer.Reload;
var
  S, T: PSItem;
  I: Integer;
begin
  with Listing^ do
  begin
    T := nil;
    for i := Fields^.Count - 1 downto 0 do
    begin
      S := NewSItem(PString(Fields^.At(i))^, T);
      T := S;
    end;
    MakeList(Base, S);
  end;
end;

{ *** TDataEditor object *** }

constructor TDataEditor.Init(var R: TRect; PDbf: PBase; PFields: PSItem);
begin
  inherited Init(R, PDbf, PFields);
  with EdLine^ do Options := Options or ofSelectable;
end;

procedure TDataEditor.AutoSave;
var
  ChkEof: Boolean;
  Tmp: PCollection;
begin
  with Listing^, EdLine^ do
  begin
    if (Original = nil) and (Data^ = '') or
      (Original <> nil) and (Original^ = Data^) then Exit;
    ChkEof := Base^.Eof;
    if ChkEof then
    begin
      if Data^ = '' then Exit;
      Base^.Append;
    end;
    Base^.Put(PString(Fields^.At(Column))^, EdLine^.Data^);
    if ChkEof then
    begin
      List^.AtInsert(Focused, NewLine);
      Tmp := List;  List := nil;
      NewList(Tmp);
      FocusItem(Range - 2);
    end else
      List^.AtPut(Focused, NewLine);
  end;
end;

procedure TDataEditor.HandleEvent(var Event: Tevent);
var
  S: string;
begin
  inherited HandleEvent(Event);
  if Event.What = evKeyDown then
  case Event.KeyCode of
    kbDel:
      begin
        if (EdLine^.State and sfFocused <> sfFocused) and
          (MessageBox(^C' / ⠭ ⥪   ?',
          nil, mfConfirmation + mfYesNoCancel) = cmYes) then
        with Listing^.Base^ do
        begin
          if Deleted then Recall else Delete;
          Self.Redraw;
        end;
        ClearEvent(Event);
      end;
    kbEsc:
      if EdLine^.State and sfFocused = sfFocused then
      begin
        with EdLine^ do
        begin
          if Original <> nil then S := Original^ else S := '';
          SetData(S);
        end;
        Listing^.Select;
        ClearEvent(Event);
      end;
    kbEnter, kbCtrlEnter:
      begin
        if EdLine^.State and sfFocused <> sfFocused then
        begin
          EdLine^.Select;
          ClearEvent(Event);
          Exit;
        end;
        Lock;
        with Listing^ do
        begin
          AutoSave;
          if Event.KeyCode = kbCtrlEnter then JumpVert := not JumpVert;
          case JumpVert of
            False:  { ਧ⠫쭮 ६饭 }
              begin
                if Column < Fields^.Count - 1 then Inc(Column) else
                  if Focused <> Range - 1 then
                  begin
                    Column := 0;
                    FocusItem(Focused + 1);
                  end;
                HScroll^.SetValue(Column);
              end;
            True:  { ⨪쭮 ६饭 }
              if Focused <> Range - 1 then FocusItem(Focused + 1);
          end;
        end; {with}
        UnLock;
        ClearEvent(Event);
      end; {kbEnter}
  end; {case}
end; {proc}

{ *** TCellEditor object *** }

constructor TCellEditor.Init(var R: TRect; PDbf: PBase; PFields: PSItem);
begin
  inherited Init(R, PDbf, PFields);
  Options := Options or ofPreProcess;
end;

procedure TCellEditor.HandleEvent(var Event: TEvent);
begin
  if Event.What = evKeyDown then
    if Event.CharCode in [#32..#255] then
      EdLine^.Select
    else
      case Event.KeyCode of
        kbUp, kbDown, kbLeft, kbRight,
        kbHome, kbEnd, kbPgUp, kbPgDn,
        kbCtrlLeft, kbCtrlRight,
        kbCtrlPgUp, kbCtrlPgDn:
          if EdLine^.State and sfFocused = sfFocused then
          begin
            Lock;
            AutoSave;
            Listing^.Select;
            UnLock;
          end;
      end; {case}
  inherited HandleEvent(Event);
end;

end.